import { Hono } from "hono";
import { routes } from "./routes/index.ts";
import {
    Constants,
    Innertube,
    Log,
    Platform,
    Player,
    UniversalCache,
} from "youtubei.js";
import { poTokenGenerate } from "./lib/jobs/potoken.ts";
import { konfigLoader } from "./lib/helpers/konfigLoader.ts";
import { ICache } from "https://raw.githubusercontent.com/LuanRT/YouTube.js/refs/tags/v13.0.0-deno/deno/src/types/Cache.ts";
import { FetchFunction } from "https://raw.githubusercontent.com/LuanRT/YouTube.js/refs/tags/v13.0.0-deno/deno/src/types/PlatformShim.ts";
import { getRandomUserAgent, PlayerError } from "youtubei.js/Utils";
let getFetchClientLocation = "getFetchClient";
if (Deno.env.get("GET_FETCH_CLIENT_LOCATION")) {
    if (Deno.env.has("DENO_COMPILED")) {
        getFetchClientLocation = Deno.mainModule.replace("src/main.ts", "") +
            Deno.env.get("GET_FETCH_CLIENT_LOCATION");
    } else {
        getFetchClientLocation = Deno.env.get(
            "GET_FETCH_CLIENT_LOCATION",
        ) as string;
    }
}
const { getFetchClient } = await import(getFetchClientLocation);

const app = new Hono();
const konfigStore = await konfigLoader();

let innertubeClient: Innertube;
let innertubeClientFetchPlayer = true;
const innertubeClientOauthEnabled = konfigStore.get(
    "youtube_session.oauth_enabled",
) as boolean;
const innertubeClientJobPoTokenEnabled = konfigStore.get(
    "jobs.youtube_session.po_token_enabled",
) as boolean;
const innertubeClientCookies = konfigStore.get(
    "jobs.youtube_session.cookies",
) as string;
let innertubeClientCache = new UniversalCache(
    true,
    konfigStore.get("cache.directory") as string + "/youtubei.js/",
) as UniversalCache;

Deno.env.set("TMPDIR", konfigStore.get("cache.directory") as string);

if (!innertubeClientOauthEnabled) {
    if (innertubeClientJobPoTokenEnabled) {
        console.log("[INFO] job po_token is active.");
        // Don't fetch fetch player yet for po_token
        innertubeClientFetchPlayer = false;
    } else if (!innertubeClientJobPoTokenEnabled) {
        console.log("[INFO] job po_token is NOT active.");
    }
} else if (innertubeClientOauthEnabled) {
    // Can't use cache if using OAuth#cacheCredentials
    innertubeClientCache = new UniversalCache(false);
}

innertubeClient = await Innertube.create({
    cache: innertubeClientCache,
    retrieve_player: innertubeClientFetchPlayer,
    fetch: getFetchClient(konfigStore),
    cookie: innertubeClientCookies || undefined,
});

const TAG = "Player";

export default class CustomPlayer extends Player {
    constructor(
        signature_timestamp: number,
        sig_sc?: string,
        nsig_sc?: string,
    ) {
        super("7795af42", signature_timestamp, sig_sc, nsig_sc);
    }

    static override async create(
        cache: ICache | undefined,
        fetch: FetchFunction = Platform.shim.fetch,
        po_token?: string,
    ): Promise<CustomPlayer> {
        const url = new URL("/iframe_api", Constants.URLS.YT_BASE);
        const res = await fetch(url);

        if (!res.ok) {
            throw new PlayerError(
                `Failed to get player id: ${res.status} (${res.statusText})`,
            );
        }

        const player_id = "7795af42";

        Log.info(
            TAG,
            `Got player id (${player_id}). Checking for cached players..`,
        );

        if (!player_id) {
            throw new PlayerError("Failed to get player id");
        }

        // We have the player id, now we can check if we have a cached player.
        if (cache) {
            const cached_player = await Player.fromCache(cache, player_id);
            if (cached_player) {
                Log.info(TAG, "Found up-to-date player data in cache.");
                cached_player.po_token = po_token;
                return cached_player;
            }
        }

        const player_url = new URL(
            `/s/player/${player_id}/player_ias.vflset/en_US/base.js`,
            Constants.URLS.YT_BASE,
        );

        Log.info(
            TAG,
            `Could not find any cached player. Will download a new player from ${player_url}.`,
        );

        const player_res = await fetch(player_url, {
            headers: {
                "user-agent": getRandomUserAgent("desktop"),
            },
        });

        if (!player_res.ok) {
            throw new PlayerError(
                `Failed to get player data: ${player_res.status}`,
            );
        }

        const player_js = await player_res.text();

        const sig_timestamp = this.extractSigTimestamp(player_js);
        const sig_sc = this.extractSigSourceCode(player_js);
        const nsig_sc = this.extractNSigSourceCode(player_js);

        Log.info(
            TAG,
            `Got signature timestamp (${sig_timestamp}) and algorithms needed to decipher signatures.`,
        );

        const player = await Player.fromSource(
            player_id,
            sig_timestamp,
            cache,
            sig_sc,
            nsig_sc,
        );
        player.po_token = po_token;

        return player as CustomPlayer;
    }
}

export const player = await CustomPlayer.create(undefined);

innertubeClient.session.player = player;

const poTokenRefreshInterval = Deno.env.get("FREQUENCY_SECONDS") ||
    konfigStore.get("jobs.youtube_session.frequency_seconds");
console.log("[INFO] po_token refresh interval set to", poTokenRefreshInterval);

if (!innertubeClientOauthEnabled) {
    if (innertubeClientJobPoTokenEnabled) {
        innertubeClient = await poTokenGenerate(
            innertubeClient,
            konfigStore,
            innertubeClientCache as UniversalCache,
        );
    }
    setInterval(
        async () => {
            if (innertubeClientJobPoTokenEnabled) {
                innertubeClient = await poTokenGenerate(
                    innertubeClient,
                    konfigStore,
                    innertubeClientCache,
                );
            } else {
                innertubeClient = await Innertube.create({
                    cache: innertubeClientCache,
                    retrieve_player: innertubeClientFetchPlayer,
                });
            }
        },
        poTokenRefreshInterval as number * 1000 || 50000,
    );
} else if (innertubeClientOauthEnabled) {
    // Fired when waiting for the user to authorize the sign in attempt.
    innertubeClient.session.on("auth-pending", (data) => {
        console.log(
            `[INFO] [OAUTH] Go to ${data.verification_url} in your browser and enter code ${data.user_code} to authenticate.`,
        );
    });
    // Fired when authentication is successful.
    innertubeClient.session.on("auth", () => {
        console.log("[INFO] [OAUTH] Sign in successful!");
    });
    // Fired when the access token expires.
    innertubeClient.session.on("update-credentials", async () => {
        console.log("[INFO] [OAUTH] Credentials updated.");
        await innertubeClient.session.oauth.cacheCredentials();
    });

    // Attempt to sign in and then cache the credentials
    await innertubeClient.session.signIn();
    await innertubeClient.session.oauth.cacheCredentials();
}

app.use("*", async (c, next) => {
    // @ts-ignore Do not understand how to fix this error.
    c.set("innertubeClient", innertubeClient);
    // @ts-ignore Do not understand how to fix this error.
    c.set("konfigStore", konfigStore);
    await next();
});

routes(app, konfigStore);

const https = Deno.env.get("HTTPS");
const port = konfigStore.get("server.port") as number;
const host = konfigStore.get("server.host") as string;

if (https == "TRUE" || https == "true") {
    const cert = Deno.readTextFileSync("/data/cert.pem");
    const key = Deno.readTextFileSync("/data/key.key");
    Deno.serve({
        port: port,
        hostname: host,
        cert: cert,
        key: key,
    }, app.fetch);
} else {
    Deno.serve({
        port: port,
        hostname: host,
    }, app.fetch);
}
